@@ -31,11 +31,11 @@
 #include "moduleloader.h"
 
 #include "builtindeclarations.h"
-#include "builtinvalue.h"
 #include "evaluator.h"
 #include "filecontext.h"
 #include "item.h"
 #include "itemreader.h"
+#include "language.h"
 #include "modulemerger.h"
 #include "qualifiedid.h"
 #include "scriptengine.h"
@@ -102,6 +102,13 @@ public:
         QSet<ProductContext *> rootProducts = allProducts.toSet() - allDependencies;
         foreach (ProductContext * const rootProduct, rootProducts)
             traverse(rootProduct);
+        if (m_sortedProducts.count() < allProducts.count()) {
+            foreach (auto * const product, allProducts) {
+                QList<ModuleLoader::ProductContext *> path;
+                findCycle(product, path);
+            }
+        }
+        QBS_CHECK(m_sortedProducts.count() == allProducts.count());
     }
 
     // No product at position i has dependencies to a product at position j > i.
@@ -121,6 +128,22 @@ private:
         m_sortedProducts << product;
     }
 
+    void findCycle(ModuleLoader::ProductContext *product,
+                   QList<ModuleLoader::ProductContext *> &path)
+    {
+        if (path.contains(product)) {
+            ErrorInfo error(Tr::tr("Cyclic dependencies detected."));
+            foreach (const auto * const p, path)
+                error.append(p->name, p->item->location());
+            error.append(product->name, product->item->location());
+            throw error;
+        }
+        path << product;
+        foreach (auto * const dep, m_dependencyMap.value(product))
+            findCycle(dep, path);
+        path.removeLast();
+    }
+
     TopLevelProjectContext &m_tlp;
     QHash<ProductContext *, QVector<ProductContext *>> m_dependencyMap;
     QSet<ProductContext *> m_seenProducts;
@@ -183,6 +206,11 @@ void ModuleLoader::setSearchPaths(const QStringList &searchPaths)
     }
 }
 
+void ModuleLoader::setOldProbes(const QHash<QString, QList<ProbeConstPtr>> &oldProbes)
+{
+    m_oldProbes = oldProbes;
+}
+
 ModuleLoaderResult ModuleLoader::load(const SetupProjectParameters &parameters)
 {
     if (m_logger.traceEnabled())
@@ -211,8 +239,7 @@ ModuleLoaderResult ModuleLoader::load(const SetupProjectParameters &parameters)
     root = wrapInProjectIfNecessary(root);
 
     const QString buildDirectory = TopLevelProject::deriveBuildDirectory(parameters.buildRoot(),
-            TopLevelProject::deriveId(parameters.topLevelProfile(),
-                                      parameters.finalBuildConfigurationTree()));
+            TopLevelProject::deriveId(parameters.finalBuildConfigurationTree()));
     root->setProperty(QLatin1String("sourceDirectory"),
                       VariantValue::create(QFileInfo(root->file()->filePath()).absolutePath()));
     root->setProperty(QLatin1String("buildDirectory"), VariantValue::create(buildDirectory));
@@ -320,7 +347,6 @@ private:
     }
 
     void handle(VariantValue *) { /* only created internally - no need to check */ }
-    void handle(BuiltinValue *) { /* only created internally - no need to check */ }
 };
 
 void ModuleLoader::handleTopLevelProject(ModuleLoaderResult *loadResult, Item *projectItem,
@@ -342,9 +368,8 @@ void ModuleLoader::handleTopLevelProject(ModuleLoaderResult *loadResult, Item *p
                     throw err;
                 }
                 m_logger.printWarning(err);
-                ModuleLoaderResult::ProductInfo pi;
-                pi.hasError = true;
-                it->project->result->productInfos.insert(it->item, pi);
+                it->info.hasError = true;
+                it->project->result->productInfos.insert(it->item, it->info);
                 m_disabledItems << it->item;
             }
         }
@@ -356,7 +381,6 @@ void ModuleLoader::handleTopLevelProject(ModuleLoaderResult *loadResult, Item *p
         handleProduct(p);
 
     m_reader->clearExtraSearchPathsStack();
-    checkItemTypes(projectItem);
     PropertyDeclarationCheck check(m_disabledItems, m_parameters, m_logger);
     check(projectItem);
 }
@@ -376,7 +400,7 @@ void ModuleLoader::handleProject(ModuleLoaderResult *loadResult,
     ProductContext dummyProductContext;
     dummyProductContext.project = &projectContext;
     dummyProductContext.moduleProperties = m_parameters.finalBuildConfigurationTree();
-    loadBaseModule(&dummyProductContext, projectItem);
+    projectItem->addModule(loadBaseModule(&dummyProductContext, projectItem));
     overrideItemProperties(projectItem, QLatin1String("project"),
                            m_parameters.overriddenValuesTree());
     const QString projectName = m_evaluator->stringValue(projectItem, QLatin1String("name"));
@@ -503,7 +527,7 @@ QList<Item *> ModuleLoader::multiplexProductItem(ProductContext *dummyContext, I
     ValuePtr qbsValue = productItem->property(qbsKey); // Retrieve now to restore later.
     if (qbsValue)
         qbsValue = qbsValue->clone();
-    loadBaseModule(dummyContext, productItem);
+    productItem->addModule(loadBaseModule(dummyContext, productItem));
 
     // Overriding the product item properties must be done here already, because otherwise
     // the "profiles" property would not be overridable.
@@ -573,10 +597,9 @@ void ModuleLoader::prepareProduct(ProjectContext *projectContext, Item *productI
     if (it == projectContext->result->profileConfigs.constEnd()) {
         const QVariantMap buildConfig = SetupProjectParameters::expandedBuildConfiguration(
                     m_parameters.settingsDirectory(), productContext.profileName,
-                    m_parameters.buildVariant());
+                    m_parameters.configurationName());
         productContext.moduleProperties = SetupProjectParameters::finalBuildConfigurationTree(
-                    buildConfig, m_parameters.overriddenValues(), m_parameters.buildRoot(),
-                    m_parameters.topLevelProfile());
+                    buildConfig, m_parameters.overriddenValues(), m_parameters.buildRoot());
         projectContext->result->profileConfigs.insert(productContext.profileName,
                                                       productContext.moduleProperties);
     } else {
@@ -637,6 +660,9 @@ static void createSortedModuleList(const Item::Module &parentModule, QVector<Ite
 
 void ModuleLoader::handleProduct(ModuleLoader::ProductContext *productContext)
 {
+    if (productContext->info.hasError)
+        return;
+
     Item * const item = productContext->item;
 
     Item::Modules mergedModules;
@@ -657,7 +683,32 @@ void ModuleLoader::handleProduct(ModuleLoader::ProductContext *productContext)
         if (!module.item->isPresentModule())
             continue;
         try {
-            resolveProbes(module.item);
+            resolveProbes(productContext, module.item);
+            if (module.versionRange.minimum.isValid()
+                    || module.versionRange.maximum.isValid()) {
+                if (module.versionRange.maximum.isValid()
+                        && module.versionRange.minimum >= module.versionRange.maximum) {
+                    throw ErrorInfo(Tr::tr("Impossible version constraint [%1,%2) set for module "
+                                           "'%3'").arg(module.versionRange.minimum.toString(),
+                                                       module.versionRange.maximum.toString(),
+                                                       module.name.toString()));
+                }
+                const Version moduleVersion = Version::fromString(
+                            m_evaluator->stringValue(module.item, QLatin1String("version")));
+                if (moduleVersion < module.versionRange.minimum) {
+                    throw ErrorInfo(Tr::tr("Module '%1' has version %2, but it needs to be "
+                            "at least %3.").arg(module.name.toString(),
+                                                moduleVersion.toString(),
+                                                module.versionRange.minimum.toString()));
+                }
+                if (module.versionRange.maximum.isValid()
+                        && moduleVersion >= module.versionRange.maximum) {
+                    throw ErrorInfo(Tr::tr("Module '%1' has version %2, but it needs to be "
+                            "lower than %3.").arg(module.name.toString(),
+                                               moduleVersion.toString(),
+                                               module.versionRange.maximum.toString()));
+                }
+            }
             m_evaluator->boolValue(module.item, QLatin1String("validate"));
         } catch (const ErrorInfo &error) {
             if (module.required) { // Error will be thrown for enabled products only
@@ -669,7 +720,7 @@ void ModuleLoader::handleProduct(ModuleLoader::ProductContext *productContext)
         }
     }
 
-    resolveProbes(item);
+    resolveProbes(productContext, item);
 
     if (!checkItemCondition(item)) {
         Item * const productModule = m_productModuleCache.value(productContext->name);
@@ -683,6 +734,7 @@ void ModuleLoader::handleProduct(ModuleLoader::ProductContext *productContext)
         if (child->type() == ItemType::Group)
             handleGroup(productContext, child);
     }
+    productContext->project->result->productInfos.insert(item, productContext->info);
 }
 
 void ModuleLoader::initProductProperties(const ProductContext &product)
@@ -793,6 +845,37 @@ bool ModuleLoader::checkExportItemCondition(Item *exportItem, const ProductConte
     return checkItemCondition(exportItem);
 }
 
+ProbeConstPtr ModuleLoader::findOldProbe(const QString &product,
+                                         bool condition,
+                                         const QVariantMap &initialProperties,
+                                         const QString &sourceCode) const
+{
+    if (m_parameters.forceProbeExecution())
+        return ProbeConstPtr();
+    foreach (const ProbeConstPtr &oldProbe, m_oldProbes.value(product)) {
+        if (oldProbe->condition() != condition)
+            continue;
+        if (oldProbe->configureScript() != sourceCode)
+            continue;
+        if (oldProbe->initialProperties() != initialProperties)
+            continue;
+        return oldProbe;
+    }
+
+    return ProbeConstPtr();
+}
+
+ProbeConstPtr ModuleLoader::findCurrentProbe(const CodeLocation &location, bool condition,
+                                             const QVariantMap &initialProperties) const
+{
+    const QList<ProbeConstPtr> cachedProbes = m_currentProbes.value(location);
+    foreach (const ProbeConstPtr &p, cachedProbes) {
+        if (p->condition() == condition && p->initialProperties() == initialProperties)
+            return p;
+    }
+    return ProbeConstPtr();
+}
+
 void ModuleLoader::mergeExportItems(const ProductContext &productContext)
 {
     QVector<Item *> exportItems;
@@ -825,6 +908,17 @@ void ModuleLoader::mergeExportItems(const ProductContext &productContext)
         filesWithExportItem += exportItem->file();
         foreach (Item *child, exportItem->children())
             Item::addChild(merged, child);
+        const Item::PropertyDeclarationMap &decls = exportItem->propertyDeclarations();
+        for (auto it = decls.constBegin(); it != decls.constEnd(); ++it) {
+            const PropertyDeclaration &newDecl = it.value();
+            const PropertyDeclaration &existingDecl = merged->propertyDeclaration(it.key());
+            if (existingDecl.isValid() && existingDecl.type() != newDecl.type()) {
+                ErrorInfo error(Tr::tr("Export item in inherited item redeclares property "
+                        "'%1' with different type.").arg(it.key()), exportItem->location());
+                handlePropertyError(error, m_parameters, m_logger);
+            }
+            merged->setPropertyDeclaration(newDecl.name(), newDecl);
+        }
         for (QMap<QString, ValuePtr>::const_iterator it = exportItem->properties().constBegin();
                 it != exportItem->properties().constEnd(); ++it) {
             mergeProperty(merged, it.key(), it.value());
@@ -867,8 +961,7 @@ void ModuleLoader::propagateModulesFromProduct(ProductContext *productContext, I
 
 void ModuleLoader::resolveDependencies(DependsContext *dependsContext, Item *item)
 {
-    loadBaseModule(dependsContext->product, item);
-
+    const Item::Module baseModule = loadBaseModule(dependsContext->product, item);
     // Resolve all Depends items.
     ItemModuleList loadedModules;
     ProductDependencyResults productDependencies;
@@ -876,6 +969,7 @@ void ModuleLoader::resolveDependencies(DependsContext *dependsContext, Item *ite
         if (child->type() == ItemType::Depends)
             resolveDependsItem(dependsContext, item, child, &loadedModules, &productDependencies);
 
+    item->addModule(baseModule);
     foreach (const Item::Module &module, loadedModules)
         item->addModule(module);
 
@@ -968,6 +1062,12 @@ void ModuleLoader::resolveDependsItem(DependsContext *dependsContext, Item *pare
             if (!m_requiredChain.at(i))
                 isRequired = false;
         }
+        const Version minVersion = Version::fromString(
+                    m_evaluator->stringValue(dependsItem, QLatin1String("versionAtLeast")));
+        const Version maxVersion = Version::fromString(
+                    m_evaluator->stringValue(dependsItem, QLatin1String("versionBelow")));
+        const VersionRange versionRange(minVersion, maxVersion);
+
         // Don't load the same module twice. Duplicate Depends statements can easily
         // happen due to inheritance.
         const auto it = std::find_if(moduleResults->begin(), moduleResults->end(),
@@ -975,6 +1075,7 @@ void ModuleLoader::resolveDependsItem(DependsContext *dependsContext, Item *pare
         if (it != moduleResults->end()) {
             if (isRequired)
                 it->required = true;
+            it->versionRange.narrowDown(versionRange);
             continue;
         }
 
@@ -983,8 +1084,8 @@ void ModuleLoader::resolveDependsItem(DependsContext *dependsContext, Item *pare
         Item *moduleItem = loadModule(dependsContext->product, parentItem, dependsItem->location(),
                                       dependsItem->id(), moduleName, isRequired, &result.isProduct);
         if (!moduleItem) {
-            // ### 1.5: change error message to the more generic "Dependency '%1' not found.".
-            throw ErrorInfo(Tr::tr("Product dependency '%1' not found.").arg(moduleName.toString()),
+            throw ErrorInfo(Tr::tr("Dependency '%1' not found for product '%2'.")
+                            .arg(moduleName.toString(), dependsContext->product->name),
                             dependsItem->location());
         }
         if (m_logger.traceEnabled())
@@ -992,6 +1093,7 @@ void ModuleLoader::resolveDependsItem(DependsContext *dependsContext, Item *pare
         result.name = moduleName;
         result.item = moduleItem;
         result.required = isRequired;
+        result.versionRange = versionRange;
         moduleResults->append(result);
         if (result.isProduct) {
             if (m_logger.traceEnabled())
@@ -1281,7 +1383,7 @@ Item *ModuleLoader::loadModuleFile(ProductContext *productContext, const QString
     return module;
 }
 
-void ModuleLoader::loadBaseModule(ProductContext *productContext, Item *item)
+Item::Module ModuleLoader::loadBaseModule(ProductContext *productContext, Item *item)
 {
     const QualifiedId baseModuleName(QLatin1String("qbs"));
     Item::Module baseModuleDesc;
@@ -1291,7 +1393,7 @@ void ModuleLoader::loadBaseModule(ProductContext *productContext, Item *item)
     QBS_CHECK(!baseModuleDesc.isProduct);
     if (Q_UNLIKELY(!baseModuleDesc.item))
         throw ErrorInfo(Tr::tr("Cannot load base qbs module."));
-    item->addModule(baseModuleDesc);
+    return baseModuleDesc;
 }
 
 static QStringList hostOS()
@@ -1421,11 +1523,6 @@ static QStringList hostOS()
 
 void ModuleLoader::setupBaseModulePrototype(Item *prototype)
 {
-    prototype->setProperty(QLatin1String("getEnv"),
-                           BuiltinValue::create(BuiltinValue::GetEnvFunction));
-    prototype->setProperty(QLatin1String("currentEnv"),
-                           BuiltinValue::create(BuiltinValue::CurrentEnvFunction));
-
     prototype->setProperty(QLatin1String("hostOS"), VariantValue::create(hostOS()));
     prototype->setProperty(QLatin1String("libexecPath"),
                            VariantValue::create(m_parameters.libexecPath()));
@@ -1611,14 +1708,14 @@ void ModuleLoader::createChildInstances(ProductContext *productContext, Item *in
     }
 }
 
-void ModuleLoader::resolveProbes(Item *item)
+void ModuleLoader::resolveProbes(ProductContext *productContext, Item *item)
 {
     foreach (Item *child, item->children())
         if (child->type() == ItemType::Probe)
-            resolveProbe(item, child);
+            resolveProbe(productContext, item, child);
 }
 
-void ModuleLoader::resolveProbe(Item *parent, Item *probe)
+void ModuleLoader::resolveProbe(ProductContext *productContext, Item *parent, Item *probe)
 {
     m_logger.qbsTrace() << "Resolving Probe at " << probe->location().toString();
     const JSSourceValueConstPtr configureScript = probe->sourceProperty(QLatin1String("configure"));
@@ -1626,11 +1723,15 @@ void ModuleLoader::resolveProbe(Item *parent, Item *probe)
         throw ErrorInfo(Tr::tr("Probe.configure must be set."), probe->location());
     typedef QPair<QString, QScriptValue> ProbeProperty;
     QList<ProbeProperty> probeBindings;
+    QVariantMap initialProperties;
     for (Item *obj = probe; obj; obj = obj->prototype()) {
         foreach (const QString &name, obj->properties().keys()) {
             if (name == QLatin1String("configure"))
                 continue;
-            probeBindings += ProbeProperty(name, m_evaluator->value(probe, name));
+            const QScriptValue value = m_evaluator->value(probe, name);
+            probeBindings += ProbeProperty(name, value);
+            if (name != QLatin1String("condition"))
+                initialProperties.insert(name, value.toVariant());
         }
     }
     QScriptValue scope = m_engine->newObject();
@@ -1639,20 +1740,34 @@ void ModuleLoader::resolveProbe(Item *parent, Item *probe)
     m_engine->currentContext()->pushScope(scope);
     foreach (const ProbeProperty &b, probeBindings)
         scope.setProperty(b.first, b.second);
-    const bool runProbe = m_evaluator->boolValue(probe, QLatin1String("condition"));
+    const bool condition = m_evaluator->boolValue(probe, QLatin1String("condition"));
+    ProbeConstPtr resolvedProbe = findOldProbe(productContext->name, condition, initialProperties,
+                                               configureScript->sourceCode().toString());
+    if (!resolvedProbe)
+        resolvedProbe = findCurrentProbe(probe->location(), condition, initialProperties);
     ErrorInfo evalError;
-    if (runProbe) {
+    if (!condition) {
+        m_logger.qbsDebug() << "Probe disabled; skipping";
+    } else if (!resolvedProbe) {
         QScriptValue sv = m_engine->evaluate(configureScript->sourceCodeForEvaluation());
         if (Q_UNLIKELY(m_engine->hasErrorOrException(sv)))
             evalError = ErrorInfo(m_engine->lastErrorString(sv), configureScript->location());
-    } else {
-        m_logger.qbsDebug() << "Probe disabled; skipping";
     }
+    QVariantMap properties;
     foreach (const ProbeProperty &b, probeBindings) {
-        const QVariant newValue = scope.property(b.first).toVariant();
+        const QVariant newValue = resolvedProbe
+                ? resolvedProbe->properties().value(b.first) : scope.property(b.first).toVariant();
         if (newValue != b.second.toVariant())
             probe->setProperty(b.first, VariantValue::create(newValue));
+        if (!resolvedProbe)
+            properties.insert(b.first, newValue);
     }
+    if (!resolvedProbe) {
+        resolvedProbe = Probe::create(probe->location(), condition,
+                configureScript->sourceCode().toString(), properties, initialProperties);
+        m_currentProbes[probe->location()] << resolvedProbe;
+    }
+    productContext->info.probes << resolvedProbe;
     m_engine->currentContext()->popScope();
     m_engine->currentContext()->popScope();
     m_engine->currentContext()->popScope();
@@ -1664,8 +1779,7 @@ void ModuleLoader::checkCancelation() const
 {
     if (m_progressObserver && m_progressObserver->canceled()) {
         throw ErrorInfo(Tr::tr("Project resolving canceled for configuration %1.")
-                    .arg(TopLevelProject::deriveId(m_parameters.topLevelProfile(),
-                                                   m_parameters.finalBuildConfigurationTree())));
+                    .arg(TopLevelProject::deriveId(m_parameters.finalBuildConfigurationTree())));
     }
 }
 
@@ -1677,22 +1791,6 @@ bool ModuleLoader::checkItemCondition(Item *item)
     return false;
 }
 
-void ModuleLoader::checkItemTypes(Item *item)
-{
-    const ItemDeclaration decl = BuiltinDeclarations::instance().declarationsForType(item->type());
-    foreach (Item *child, item->children()) {
-        if (child->type() > ItemType::LastActualItem)
-            continue;
-        checkItemTypes(child);
-        if (!decl.isChildTypeAllowed(child->type()))
-            throw ErrorInfo(Tr::tr("Items of type '%1' cannot contain items of type '%2'.")
-                .arg(item->typeName(), child->typeName()), item->location());
-    }
-
-    foreach (const Item::Module &m, item->modules())
-        checkItemTypes(m.item);
-}
-
 QStringList ModuleLoader::readExtraSearchPaths(Item *item, bool *wasSet)
 {
     QStringList result;
@@ -1810,6 +1908,7 @@ static void collectAllModules(Item *item, QVector<Item::Module> *modules)
             // If a module is required somewhere, it is required in the top-level item.
             if (m.required)
                 it->required = true;
+            it->versionRange.narrowDown(m.versionRange);
             continue;
         }
         modules->append(m);
@@ -1859,6 +1958,7 @@ void ModuleLoader::addTransitiveDependencies(ProductContext *ctx)
             }
             dep.name = module.name;
             dep.required = module.required;
+            dep.versionRange = module.versionRange;
             ctx->item->addModule(dep);
         }
     }
